=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# © Copyright Jan 2014
# Designed and developped Jan 2014 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   bootstrap_VisuHole.rb
# Original Date	:   28 Jan 2014
# Description	:   Loader for the VisuHole menu and toolbar
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_VisuHole

#---------------------------------------------------------------------------------------------------------------------------
# Overall Configuration for the Tools	
#---------------------------------------------------------------------------------------------------------------------------

MENU_PERSO = [:T_MenuTools_Fredo6Collection, "Tools", :T_MenuPlugins_Fredo6Collection, "Plugins"]

#---------------------------------------------------------------------------------------------------------------------------
#Constants for TOS  Modules (do not translate here, use Translation Dialog Box instead)	
#---------------------------------------------------------------------------------------------------------------------------

#Menu and Toolbar icons
T6[:VHL_Stencil_Menu] = "VisuHole: Stencil"
T6[:VHL_Stencil_Tooltip] = "Stencil: Variations around Holes and Carving based on a stencil shape"
VHL_Stencil_Icon = "Stencil"

#--------------------------------------------------------
# Default Parameters
#--------------------------------------------------------

T6[:VHL_OPTION_HistoryKeep] = "Number of stencils kept in History"
T6[:VHL_OPTION_DeviationSmooth] = "Maximum deviation to smooth position of a curved surface - 0 to disable smoothing (angle in degrees)"
T6[:VHL_OPTION_DeviationStop] = "Deviation from which shapes are not drawn on a surface - 180 to ignore (angle in degrees)"

#---------------------------------------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------
# EXTERNAL API to launch RoundCorner menus
# Action code must be passed as a symbol
#---------------------------------------------------------------------------------------------------------------------------------
#---------------------------------------------------------------------------------------------------------------------------------
	
def F6_VisuHole.launch_action(action_code, *args)
	MYPLUGIN.launch_action action_code, *args
end

#--------------------------------------------------------------
# Condition for contextual menu selection
#--------------------------------------------------------------

def F6_VisuHole.selection_contextual?
	selection = Sketchup.active_model.selection
	return nil if selection.empty?
	nil
end

#--------------------------------------------------------
# Startup routine (called from LibFredo6)
#--------------------------------------------------------
	
#Create all commands with menus and toolbar icons
def F6_VisuHole.startup	

	#Conditions for contextual menus
	tc_edge = Traductor::TestCondition.new() { F6_VisuHole.selection_contextual? == :edge }
	submenu = T6[:VHL_Menu]
	
	#Top menu
	MYPLUGIN.declare_topmenu nil, MENU_PERSO

	#Declaring command icons
	MYPLUGIN.declare_separator
	MYPLUGIN.declare_command(:VHL_Stencil_) { F6_VisuHole.launch_action :stencil }
	
	#Declaring the default parameter
	F6_VisuHole.default_param
	
	#Startup of the Plugin
	MYPLUGIN.go	
end
	
#Declaring the Default Parameters	
def F6_VisuHole.default_param
	dp = MYDEFPARAM
	dp.separator :VHL_DEFAULT_SECTION_General, nil, T6[:T_DEFAULT_SECTION_General]
	dp.declare :VHL_OPTION_HistoryKeep, 15, 'I:>=5<=40'	
	dp.declare :VHL_OPTION_DeviationSmooth, 25.0, 'F:>=0<=45'	
	dp.declare :VHL_OPTION_DeviationStop, 45.0, 'F:>=15<=180'	
end
		
end	#End Module F6_VisuHole
